#include "server.h"

#include "server.h"
#include <stdio.h>
#include <string.h>
#include "main.h"

#ifndef _WIN32
char* _strdup(const char *s)
{
	if (!s) return NULL;
	size_t len = strlen(s);
	char *mem = (char*)malloc(len + 1);
	if (!mem) return NULL;
	memcpy(mem, s, len + 1);
	return mem;
}
#define sprintf_s sprintf


#include <sys/time.h>
#include <sys/socket.h>
#else
#pragma comment(lib, "ws2_32.lib")
#endif


Server server;



//------------------------ class Connection -----------------------------
Connection::Connection(SOCKET s)
{
	handle = s;
	timer = 15000;


	prev = NULL;
	next = server.connections;
	if (next) next->prev = this;
	server.connections = this;



	for(int i=SERVER_SAVE_MESSAGE_COUNT-1; i>=0; i--)
	{
		if (server.messages[i]) Send(server.messages[i]);
	}
}

Connection::~Connection()
{
	if (handle!=INVALID_SOCKET)
	{
		closesocket(handle);
		handle = INVALID_SOCKET;
		printf("Disconnected\n");
	}
	if (next) next->prev = prev;
	if (prev) prev->next = next;
	if (server.connections == this) server.connections = next;
}	

void Connection::Send(const char *text)
{
	if (!text || !text[0]) return;

	int len = (int)strlen(text);
	outData.Seek(0,2);
	outData.WriteLong(len);
	outData.Write((const unsigned char*)text, len);
}


void Connection::SendFastCommands()
{
	char buffer[20000];
	sprintf_s(buffer, "%sfastcmd%d.txt", appPath, server.port);
	FILE *f;

	#ifdef _WIN32
	if (fopen_s(&f, buffer, "rb")!=0)
	#else
	if ((f = fopen(buffer, "rb"))==NULL)
	#endif
	{
		return;
	}

	static const char fastCmd[] = "@FASTCOMMANDS\n";
	memcpy(buffer, fastCmd, sizeof(fastCmd) - 1);
	size_t n = fread(buffer + sizeof(fastCmd) - 1, 1, sizeof(buffer) - sizeof(fastCmd) - 2, f);
	if (n > 0)
	{
		n+=sizeof(fastCmd) - 1;
		buffer[n] = 0;
		Send(buffer);
	}
	fclose(f);
}


void Connection::Tick(int dT)
{
	static char buffer[0x8000];

	if (outData.size > 0)
	{
		int n = outData.size;
		if (n > 0x10000) n = 0x10000;

		n = send(handle, (const char*)outData.memory, n, 0);
		if (n > 0)
		{
			outData.Cut(n);
		}
	}


	if (outData.size == 0)
	{
		timer-=dT;
		if (timer <=0)
		{
			printf("TIMEOUT\n");
			delete this;
			return;
		}
	}

	int len = recv(handle, buffer, sizeof(buffer), 0);
	if (len == 0)
	{
		delete this;
		return;
	}
	if (len > 0)
	{
		timer = 15000;
		inData.Seek(0,2);
		inData.Write(buffer, len);
		

		while (inData.size >= 4)
		{
			inData.Seek(0,0);
			int msgSize = inData.ReadLong();

			if (msgSize < 0 || msgSize > 0x20000)
			{
				printf("Invalid msg size\n");
				delete this;
				return;
			}

			if (inData.size < msgSize + 4) return; //  

			timer = 15000;

			char *inMsg = (char*)malloc(msgSize + 1);
			if (inMsg)
			{
				inData.Read((unsigned char*)inMsg, msgSize);
				inMsg[msgSize] = 0;

				if (strcmp(inMsg,"@PING")==0)
				{
					printf("@PING\n");
					Send("@PING");
					free(inMsg);
				}
				else
				if (strcmp(inMsg,"@FASTCOMMANDS")==0)
				{
					SendFastCommands();
					free(inMsg);
				}
				else
				{
					//------- add message --------
					free(server.messages[SERVER_SAVE_MESSAGE_COUNT-1]);
					for(int i=SERVER_SAVE_MESSAGE_COUNT-1; i>=1; i--)
					{
						server.messages[i] = server.messages[i-1];
					}
					server.messages[0] = inMsg;

					//------ sending -----------
					Connection *conn = server.connections;
					while(conn)
					{
						conn->Send(inMsg);
						conn = conn->next;
					}
				}

			}
			inData.Cut(msgSize + 4);
		}
	}
}




//---------------------------- Server ----------------------------
Server::Server()
{
	handle = INVALID_SOCKET;
	connections = NULL;
	for(int i=0; i < SERVER_SAVE_MESSAGE_COUNT; i++) messages[i] = NULL;
}

Server::~Server()
{
	Close();
}

	
bool Server::Open(unsigned short port)
{
	Close();

	this->port = port;
	handle = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (handle == INVALID_SOCKET)
	{
		return false;
	}
	
	int optval = 1;
	
	setsockopt(handle, SOL_SOCKET, SO_REUSEADDR, (char*)&optval, sizeof(optval));
	

	#ifdef _WIN32
	u_long mode = 1;
	if (ioctlsocket(handle, FIONBIO, &mode)!=NO_ERROR)
	{
		closesocket(handle);
		handle = INVALID_SOCKET;
		return false;
	}
	#else
	if (fcntl(handle, F_SETFL, O_NONBLOCK)<0)
	{
		closesocket(handle);
		handle = INVALID_SOCKET;
		return false;
	}
	#endif


	
	struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = 0; //127 | (1 << 24);

	if (bind(handle, (struct sockaddr *)&addr, sizeof(addr)) < 0)
	{
    	closesocket(handle);
		handle = INVALID_SOCKET;
   		return false;
   	}


	if (listen(handle, 5) < 0)
	{
		closesocket(handle);
		handle = INVALID_SOCKET;
   		return false;
	}

	return true;
}

void Server::Close()
{
	while(connections) delete connections;
	if (handle!=INVALID_SOCKET)
	{
		closesocket(handle);
		handle = INVALID_SOCKET;
	}
}
	
void Server::Tick(int dT)
{
	if (handle != INVALID_SOCKET)
	{
		SOCKET s = accept(handle, NULL, NULL);
		if (s != INVALID_SOCKET)
		{
			#ifdef _WIN32
			u_long mode = 1;
			if (ioctlsocket(s, FIONBIO, &mode)!=NO_ERROR)
			{
				closesocket(s);
				s = INVALID_SOCKET;
			}
			#else
			if (fcntl(s, F_SETFL, O_NONBLOCK)<0)
			{
				closesocket(s);
				s = INVALID_SOCKET;
			}
			#endif

			if (s!=INVALID_SOCKET)
			{
				printf("New client connected\n");
				new Connection(s);
			}
		}
	}

	Connection *conn = connections;
	while(conn)
	{
		Connection *next = conn->next;
		conn->Tick(dT);
		conn = next;
	}
}
