#ifndef __streams__
#define __streams__
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef _WIN32
#include <tchar.h>
#else
typedef char TCHAR;
#define _T(x) x
#endif

class CStream
{
public:
	CStream();
	virtual ~CStream();
	virtual int Write(const void *mem, int count) = 0;
	virtual int Read(void *mem, int count) = 0;
	virtual int Seek(int offset, int origin) = 0;
	virtual int Size();
	virtual int Position();
	inline int Last() { return Size() - Position(); }

	bool ReadBool();
	unsigned char ReadByte();
	unsigned short ReadShort();
	int ReadLong();
	float ReadFloat();
	unsigned long long ReadUInt64();
	double ReadDouble();
	char * ReadAnsiString();

	bool WriteBool(bool val);
	bool WriteByte(unsigned char val);
	bool WriteShort(unsigned short val);
	bool WriteLong(int val);
	bool WriteUInt64(unsigned long long val);
	bool WriteFloat(float val);
	bool WriteDouble(double val);
	bool WriteString(const char * val);
	bool WriteText(const char *text);
	bool Format(const char *format,...);

	bool CopyFrom(CStream *source, int count);

	char *ReadText(char *buffer, size_t bufferSize);
	int ReadTextInt(int def=0.0);
	double ReadTextFloat(double def = 0.0);

};

class CFileStream : public CStream
{
public:
	FILE* f;
	CFileStream(const TCHAR *fileName, const TCHAR *mode);
	~CFileStream();
	int Write(const void *mem, int count);
	int Read(void *mem, int count);
	int Seek(int offset, int origin);
	void Close();
};


class CMemoryStream:public CStream
{
public:
	unsigned char *memory;
	int  position;
	int  size;
	int  blockSize; // must be power of 2

	CMemoryStream();
	~CMemoryStream();
	void Cut();
	void Cut(int pos);
	int Write(const void *mem, int count);
	int Read(void *mem, int count);
	int Seek(int offset, int origin);
	int Find(const char *str, int len=-1);
	int Find(int startPos, const char *str, int len=-1);
	int FindEnd(const char *str, int len=-1);
	int FindEnd(int startPos, const char *str, int len=-1);
	bool Insert(int pos, const char *str, int len=-1);
	void Delete(int startPos, int len=-1);
	bool SetSize(int newSize);
	bool SaveToFile(const TCHAR *fileName);
	bool LoadFromFile(const TCHAR *fileName);
};


#endif



